﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using ICSharpCode.SharpZipLib.Checksums;

namespace FileTransfer
{
    public partial class Form1 : Form
    {
        private Thread connector;
        private TcpClient client;
        private volatile NetworkStream stream;
        private IPAddress ip;
        private volatile bool connectorRunning;
        private readonly int port = 21341;
        private readonly int STREAM_PREFIX = -2;
        private string fileToSend;
        public Form1()
        {
            InitializeComponent();
            
        }
        private void Form1_Load(object sender, EventArgs e)
        {
            SetStatus(false);
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            if (connectorRunning)
            {
                log("Connector is already running.");
                return;
            }
            try
            {
                ip = IPAddress.Parse(txtIP.Text);
            }
            catch (Exception ee)
            {
                log(ee.Message);
            }
            connectorRunning = true;
            connector = new Thread(startConnection);
            connector.Name = "Connector";
            connector.IsBackground = true;
            connector.Start();
        }
        private void SetStatus(bool connected)
        {
            MethodInvoker mi = delegate()
            {
                btnSend.Enabled = connected;
                btnConnect.Enabled = !connected;
                lblStatus.Text = connected ? "Connected" : "Disconnected";
            };
            this.BeginInvoke(mi);
        }
        private void log(string msg)
        {
            MethodInvoker mi = delegate()
            {
                txtLogs.AppendText(msg);
                txtLogs.AppendText("\r\n");
                txtLogs.SelectionStart = txtLogs.Text.Length;
                txtLogs.ScrollToCaret();
            };
            if (this.InvokeRequired)
                this.BeginInvoke(mi);
            else
                mi.Invoke();
        }
        
        private void startConnection()
        {
            try
            {
                client = new TcpClient();
                client.Connect(ip, port);
                log("Connected!");
                SetStatus(true);
                stream = client.GetStream();
                BinaryReader br = new BinaryReader(stream);
                string fileName = "";
                while (connectorRunning)
                {
                    int prefix = br.ReadInt32();
                    if (prefix > 0) //regular message
                    {
                        byte[] buffer = br.ReadBytes(prefix);
                        fileName = Encoding.UTF8.GetString(buffer);
                        log("Receiving: " + fileName);
                    }
                    else if (prefix == STREAM_PREFIX) //stream message
                    {
                        long length = br.ReadInt64(); //total stream length
                        Adler32 adler = new Adler32();
                        using (FileStream fs = new FileStream(fileName, FileMode.Create))
                        {
                            byte[] buffer = new byte[8192];
                            long received = 0;
                            while (received < length)
                            {
                                long remain = length - received;
                                int count = stream.Read(buffer, 0, Math.Min(buffer.Length, (int)remain));
                                fs.Write(buffer, 0, count);
                                adler.Update(buffer, 0, count);
                                received += count;
                            }
                        }
                        long checkSum = br.ReadInt64();
                        if (checkSum != adler.Value)
                        {
                            throw new Exception("Checksum does not match.");
                        }
                        log("File received successfully.");
                    }
                    else
                    {
                        throw new Exception("Unexpected value: " + prefix);
                    }
                }

            }
            catch (Exception e)
            {
                log(e.Message);
                SetStatus(false);
                connectorRunning = false;
            }
        }

        private void btnSend_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog(this) == DialogResult.OK)
            {
                btnSend.Enabled = false;
                fileToSend = openFileDialog1.FileName;
                Thread sendThread = new Thread(sendFile);
                sendThread.Name = "SendThread";
                sendThread.IsBackground = true;
                sendThread.Start();
            }
        }
        private void sendFile()
        {
            try
            {
                Adler32 adler = new Adler32();
                log("Sending: " + fileToSend);
                using (FileStream fs = new FileStream(fileToSend, FileMode.Open))
                {
                    byte[] name = Encoding.UTF8.GetBytes(Path.GetFileName(fileToSend));
                    BinaryWriter bw = new BinaryWriter(stream);
                    //send the first message which is the file name
                    bw.Write(name.Length);
                    bw.Write(name);
                    //send the stream message
                    bw.Write(STREAM_PREFIX);
                    long size = new FileInfo(fileToSend).Length;
                    bw.Write(size);
                    long written = 0;
                    byte[] buffer = new byte[8192];
                    while (written < size)
                    {
                        int count = fs.Read(buffer, 0, Math.Min((int)(size - written), buffer.Length));
                        stream.Write(buffer, 0, count);
                        adler.Update(buffer, 0, count);
                        written += count;
                    }
                    bw.Write(adler.Value);
                    log("File sent successfully.");
                }
            }
            catch (Exception e)
            {
                log(e.Message);
            }
            MethodInvoker mi = delegate()
            {
                btnSend.Enabled = !btnConnect.Enabled;
            };
            this.BeginInvoke(mi);
        }


    }
}
